/*!
  @file rx_fltk_glcanvas.h
	
  @brief FLTKɂOpenGLEBhENX
 
  @author Makoto Fujisawa 
  @date   2011-09
*/

#ifndef _RX_FLTK_GLCANVAS_H_
#define _RX_FLTK_GLCANVAS_H_



//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <iostream>

// STL
#include <vector>
#include <string>

// FLTK
#include <FL/Fl.H>
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Round_Button.H>
#include <FL/Fl_Check_Button.H>

#include "rx_sph_commons.h"
#include "rx_fltk_widgets.h"

#include "rx_trackball.h"
#include "rx_model.h"
#include "rx_pov.h"

#include "rx_texture.h"

#include <boost/bind.hpp>
#include <boost/function.hpp>


using namespace std;

//-----------------------------------------------------------------------------
// `/萔
//-----------------------------------------------------------------------------
class rxFlWindow;
class rxParticleSystemBase;
struct rxCell;

class rxMCMeshCPU;
class rxMCMeshGPU;

#define RX_USE_GPU

#ifdef RX_USE_GPU
	#define RXSPH rxSPH_GPU
#else
	#define RXSPH rxSPH
#endif


// `tO
enum
{
	RX_IDD_PARTICLE			= 0x0001,	//!< p[eBN
	RX_IDD_VELOCITY			= 0x0002,	//!< x
	RX_IDD_NORMAL			= 0x0004,	//!< @
	RX_IDD_BBOX				= 0x0008,	//!< AABB(V~[V)
	RX_IDD_MESH_BBOX		= 0x0010,	//!< AABB(bV)
	RX_IDD_CELLS			= 0x0020,	//!< ߖTTpZ
	RX_IDD_MESH				= 0x0040,	//!< bV

	RX_IDD_SOLID			= 0x0080,	//!< ő
	RX_IDD_EDGE_ONLY		= 0x0100,	//!< GbŴ
	RX_IDD_EDGE				= 0x0200,	//!< GbW
	RX_IDD_REFRAC			= 0x0400,	//!< ܕ`

	RX_IDD_PARAMS			= 0x0800,	//!< p[^ʕ`
	RX_IDD_ANISOTROPICS		= 0x1000,	//!< ٕJ[l
};
const string RX_DRAW_STR[] = {
	"Particle",                "p", 
	"Velocity", 			   "v", 
	"Normal", 				   "n", 
	"AABB (Simulation Space)", "b", 
	"AABB (Mesh)", 			   "B", 
	"Cells", 				   "d", 
	"Mesh", 				   "m", 
	"Solid", 				   "o", 
	"Edge Only", 			   "E", 
	"Edge", 				   "e", 
	"Refrac", 				   "r", 
	"Params", 				   "P", 
	"Anisotoropics", 		   "A", 
	"-1"
};

// p[eBN`@
enum
{
	IDP_POINTSPRITE = 0, 
	IDP_POINT, 
	IDP_POINT_NONE, 

	IDP_END, 
};
const string RX_PARTICLE_DRAW[] = {
	"Point Sprite", "^1", 
	"GL_POINT", 	"^2", 
	"None", 		"^4", 
	"-1"
};


// Op`bV@
enum
{
	RX_TM_MC_CPU = 0, 
	RX_TM_MC_GPU, 
};
const string RX_TRIANGULATION_METHOD[] = {
	"Marching Cubes", 		"", 
	"Marching Cubes (GPU)", "", 
	"-1"
};

// ő̕`
enum
{
	RX_SOLID_VERTEX		= 0x0001, 
	RX_SOLID_EDGE		= 0x0002, 
	RX_SOLID_FACE		= 0x0004, 
	RX_SOLID_NORMAL		= 0x0008, 
	RX_SOLID_MOVE		= 0x0010,

	RX_SOLID_END
};
const string RX_SOLID_STR[] = {
	"Vertex", "", 
	"Edge",   "", 
	"Face",   "", 
	"Normal", "", 
	"Move",   "", 
	"-1"
};

//! SPHݒ
enum SettingMode
{
	ID_SPH_MESH = 0,	// bV
	ID_SPH_INPUT,		// p[eBNf[^o
	ID_SPH_OUTPUT,		// p[eBNf[^
	ID_SPH_MESH_OUTPUT, // bVo
	ID_SPH_INLET,		// E
	ID_SPH_VC, 
	ID_SPH_PS_TURB, 
	ID_SPH_ANISOTROPIC, // ٕJ[l

	ID_SPH_END, 
};


//-----------------------------------------------------------------------------
//! rxFlGLWindowNX - fltkɂOpenGLEBhE
//-----------------------------------------------------------------------------
class rxFlGLWindow : public Fl_Gl_Window
{
public:


protected:
	int m_iWinW;					//!< `EBhE̕
	int m_iWinH;					//!< `EBhE̍
	int m_iMouseButton;				//!< }EX{^̏
	int m_iKeyMod;					//!< CL[̏
	rxTrackball m_tbView;			//!< gbN{[

	double m_fBGColor[3];			//!< wiF
	bool m_bAnimation;				//!< Aj[VON/OFF

	rxFlWindow *m_pParent;			//!< eNX

	vector<rxPolygons> m_vPolys;	//!< |SIuWFNg

	// FTGL
	unsigned long m_ulFontSize;		//!< tHgTCY

	//
	// q@֘Aϐ
	//
	rxParticleSystemBase *m_pPS;	//!< SPH
	double m_fDt;					//!< ^CXebv
	double m_fGravity;				//!< d͉x

	int m_iCurrentStep;				//!< ݂̃Xebv
	bool m_bPause;					//!< V~[Ṽ|[YtO

	string m_strCurrentScene;		//!< ݂̃V[̖O
	vector<string> m_vSceneFiles;	//!< V[t@CXg
	int m_iSceneFileNum;			//!< V[t@C̐

	int m_iSimuSetting;				//!< ~[Vݒۑp


	// p[eBNo
	string m_strSphOutputName0 ;
	string m_strSphOutputHeader;

	// p[eBN
	string m_strSphInputName0 ;
	string m_strSphInputHeader;

	// ő̈ړtO
	bool m_bSolidMove;

	// ő̂̓
	Vec3 m_vMovePos[2];
	double m_fMoveMaxVel;
	bool m_bMoveSolid;
	int m_iMoveStart;


	//
	// bV
	//
	uint m_iNumVrts, m_iNumTris;	//!< ꂽbV̒_ƃbV
	int m_iVertexStore;				//!< TvO{[ɑ΂\z钸_(nx*ny*store)

	int m_iMeshMaxN;				//!< bVObh(EƂ̕)
	int m_iMeshN[3];				//!< bVObh

	Vec3 m_vMeshBoundaryExt;		//!< bVE{bNX̊eӂ̒1/2
	Vec3 m_vMeshBoundaryCen;		//!< bVE{bNX̒SW

	rxPolygons m_Poly;				//!< bV
	vector<rxPolygons> m_vSolidPoly;//!< ő̃bV
	rxMaterialOBJ m_matPoly;

	GLuint m_uVrtVBO;				//!< bV_(VBO)
	GLuint m_uTriVBO;				//!< bV|S(VBO)
	GLuint m_uNrmVBO;				//!< bV_@(VBO)
	int m_iDimVBO;

	// bVo
	int m_iSaveMeshSpacing;

	// wi摜
	bool m_bUseCubeMap;				//!< L[u}bvgptO
	rxCubeMapData m_CubeMap;		//!< L[u}bv

	// bV
	rxMCMeshCPU *m_pMCMeshCPU;
	rxMCMeshGPU *m_pMCMeshGPU;
	int m_iDepthFiltering;			//!< (fvX}bv0x01, ֊s0x02)
	int m_iSSDebugOutput;

	double m_fSpacing;				//!< fvX}bṽTvOԊu
	double m_fPrtRad;				//!< p[eBN̔a
	double m_fZmax;					//!< ֊sƂȂfvX臒l
	int m_iNfilter;					//!< fvXl̃tB^TCY
	int m_iNiters;					//!< ֊s̔

	double m_fProjectionMatrix[16];	//!< es
	double m_fModelviewMatrix[16];	//!< fr[s


public:
	// `tO
	int m_iDraw;					//!< `tO
	int m_iDrawPS;					//!< p[eBN`@
	int m_iColorType;				//!< p[eBN`掞̐F
	int m_iTriangulationMethod;		//!< Op`bV@
	int m_iSolidDraw;				//!< ő̕`

	// V~[Vݒ
	bitset<32> m_bsSimuSetting;		//!< ~[VݒtO
	double m_fVScale;				//!< xNg`掞̃XP[
	double m_fMeshThr;				//!< A֐bV臒l

	// V[Xg
	vector<string> m_vSceneTitles;	//!< V[^CgXg
	int m_iCurrentSceneIdx;			//!< ݂̃V[t@C

	// 摜o
	int m_iSaveImageSpacing;		//!< 摜ۑԊu(=-1ȂۑȂ)

public:
	//! RXgN^
	rxFlGLWindow(int x, int y, int w, int h, const char* l, void *parent);

	//! fXgN^
	~rxFlGLWindow();

public:
	void InitGL(void);
	void Resize(int w, int h);
	void Projection(void);
	vector<string> SetDrawString(void);

	// OpenGL`
	void ReDisplay(void);
	void Display(void);
	void InitView(void);

	// GUIR[obN
	void Mouse(int button, int state, int x, int y);
	void Motion(int x, int y);
	void PassiveMotion(int x, int y);
	void Idle(void);
	void Timer(void);
	void Keyboard(int key, int x, int y);
	void SpecialKey(int key, int x, int y);

	// Aj[V
	static void OnTimer_s(void* x);
	static void OnIdle_s(void* x);

	void SwitchIdle(int on);
	void SwitchFullScreen(void);

	// t@Co
	void OpenFile(const string &fn);
	void SaveFile(const string &fn);
	void SaveDisplay(const string &fn);
	void SaveDisplay(const int &stp);
	
	void SaveMesh(const string fn, rxPolygons &polys);
	void SaveMesh(const int &stp, rxPolygons &polys);

	// FTGLtHgݒ
	int SetupFonts(const char* file);


public:
	// bV
	bool CalMeshSPH(int nmax, double thr = 1000.0);
	void SetMeshCreation(bool on = true);
	RXREAL GetImplicitSPH(double x, double y, double z);

protected:
	bool calMeshSPH_CPU(int nmax, double thr = 1000.0);
	bool calMeshSPH_GPU(int nmax, double thr = 1000.0);
	bool calMeshSPH_SSM(int nmax, double thr = 1000.0);
	bool calMeshSPH_SSM_GPU(int nmax, double thr = 1000.0);


	// V~[V
	void getValueFromString(Vec3 &val, const string &str, bool rel = false, Vec3 cen = Vec3(0.0), Vec3 ext = Vec3(1.0));
	void getValueFromString(double &val, const string &str, bool rel = false, Vec3 cen = Vec3(0.0), Vec3 ext = Vec3(1.0));
	void setSphSpace(string *names, string *values, int n, string header);
	void setLiquidBox(string *names, string *values, int n, string header);
	void setLiquidSphere(string *names, string *values, int n, string header);
	void setInletLine(string *names, string *values, int n, string header);
	void SetSolidBox(string *names, string *values, int n, string header);
	void SetSolidSphere(string *names, string *values, int n, string header);
	void SetSolidPolygon(string *names, string *values, int n, string header);
	void SetMeshBoundary(string *names, string *values, int n, string header);
	void LoadSceneFromFile(const string &fn);

	void InitSPH(const string fn_scene = "null");
	void AddSphere(void);

	void StepPS(double dt);
	void ComputeFPS(void);

	void DivideString(const string &org, vector<string> &div);

	void DrawParticleVector(RXREAL *prts, RXREAL *vels, int n, int d, double *c0, double *c1, double len = 0.1);
	void DrawParticlePoints(unsigned int vbo, int n, unsigned int color_vbo = 0, RXREAL *data = 0);

	void CreateVBO(GLuint* vbo, unsigned int size);
	void DeleteVBO(GLuint* vbo);

	void DrawLiquidSurface(void);
	void SetParticleColorType(int type, int change = 0);

	void RenderSphScene(void);
	

private:
	//! `R[obN֐̃I[o[Ch
    void draw(void)
	{
		if(!context_valid()) InitGL();
		if(!valid()) Resize(w(), h());
		Display();    // OpenGL`
    }

	//! TCYR[obN֐̃I[o[Ch
    void resize(int x_, int y_, int w_, int h_)
	{
		Fl_Gl_Window::resize(x_, y_, w_, h_);
		//Resize(w_, h_);
    }


public:
	// Cxgnh
	int handle(int e);	// handle֐̃I[o[Ch

	// j[Cxgnh
	static void OnMenuDraw_s(Fl_Widget *widget, void* x);
	inline void OnMenuDraw(double val, string label);

	static void OnMenuSimulation_s(Fl_Widget *widget, void* x);
	inline void OnMenuSimulation(double val, string label);

	static void OnMenuParticle_s(Fl_Widget *widget, void* x);
	inline void OnMenuParticle(double val, string label);
	inline void OnMenuParticleColor(double val, string label);
	inline void OnMenuParticleDraw(double val, string label);

	static void OnMenuSolid_s(Fl_Widget *widget, void* x);
	inline void OnMenuSolid(double val, string label);

	static void OnMenuTriangulation_s(Fl_Widget *widget, void* x);
	inline void OnMenuTriangulation(double val, string label);

	static void OnMenuScene_s(Fl_Widget *widget, void* x);
	inline void OnMenuScene(double val, string label);
};




#endif // #ifdef _RX_FLTK_GLCANVAS_H_
