/*! 
  @file rx_sph.cu
	
  @brief CUDAɂSPH
    
  CUDPP(CUDA Data Parallel Primitives Library)Ɋւ
  http://www.gpgpu.org/developer/cudpp

  @author Makoto Fujisawa
  @date 2009-08, 2011-06
*/
// FILE --rx_sph.cu--


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cutil_inline.h>
#include <cstdio>

#include <GL/glew.h>

#include <cutil.h>
#include <cuda_gl_interop.h>

#include <GL/glut.h>

#include "rx_sph_kernel.cu"
#include "cudpp.h"



//-----------------------------------------------------------------------------
// MARK:O[oϐ
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// CUDA֐
//-----------------------------------------------------------------------------
extern "C"
{
void CuSetParameters(rxSimParams *hostParams)
{
	// copy parameters to constant memory
	cutilSafeCall( cudaMemcpyToSymbol(params, hostParams, sizeof(rxSimParams)) );
}

void CuClearData(void)
{
}


// ObhubNCubNXbȟvZ
void computeGridSize(uint n, uint blockSize, uint &numBlocks, uint &numThreads)
{
	numThreads = min(blockSize, n);
	numBlocks = DivCeil(n, numThreads);
}


//-----------------------------------------------------------------------------
// MARK:3D SPH
//-----------------------------------------------------------------------------
/*!
 * Z̃nbVvZ
 * @param[in] 
 * @return 
 */
void CuCalcHash(uint* dGridParticleHash, uint* dSortedIndex, float* dPos, int nprts)
{
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	calcHashD<<< numBlocks, numThreads >>>(dGridParticleHash,
										   dSortedIndex,
										   (float4*)dPos,
										   nprts);
	
	cutilCheckMsg("Kernel execution failed");	// J[lG[`FbN
}

/*!
 * p[eBNz\[gꂽԂɕёւC
 * eZ̎n܂ƏĨCfbNX
 * @param[in] cell p[eBNObhf[^
 * @param[in] oldPos p[eBNʒu
 * @param[in] oldVel p[eBNx
 */
void CuReorderDataAndFindCellStart(rxParticleCell cell, float* oldPos, float* oldVel)
{
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	cutilSafeCall(cudaMemset(cell.dCellStart, 0xffffffff, cell.uNumCells*sizeof(uint)));

#if USE_TEX
	cutilSafeCall(cudaBindTexture(0, dSortedPosTex, oldPos, cell.uNumParticles*sizeof(float4)));
	cutilSafeCall(cudaBindTexture(0, dSortedVelTex, oldVel, cell.uNumParticles*sizeof(float4)));
#endif

	uint smemSize = sizeof(uint)*(numThreads+1);

	// J[ls
	reorderDataAndFindCellStartD<<< numBlocks, numThreads, smemSize>>>(cell, (float4*)oldPos, (float4*)oldVel);

	cutilCheckMsg("Kernel execution failed: CuReorderDataAndFindCellStartD");
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

#if USE_TEX
	cutilSafeCall(cudaUnbindTexture(dSortedPosTex));
	cutilSafeCall(cudaUnbindTexture(dSortedVelTex));
#endif
}


/*!
 * p[eBNx̌vZ(J[lĂяo)
 * @param[out] dDens p[eBNx
 * @param[out] dPres p[eBN
 * @param[in]  cell p[eBNObhf[^
 */
void CuSphDensity(float* dDens, float* dPres, rxParticleCell cell)
{
	// MRK:CuSphDensity2D
#if USE_TEX
	cutilSafeCall(cudaBindTexture(0, dSortedPosTex, cell.dSortedPos, cell.uNumParticles*sizeof(float4)));
	cutilSafeCall(cudaBindTexture(0, dCellStartTex, cell.dCellStart, cell.uNumCells*sizeof(uint)));
	cutilSafeCall(cudaBindTexture(0, dCellEndTex, cell.dCellEnd, cell.uNumCells*sizeof(uint)));	
#endif
	//cutilSafeCall(cudaMemset((void*)dNewDens, 0, sizeof(float2)*cell.uNumParticles));

	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalDensity<<< numBlocks, numThreads >>>(dDens, dPres, cell);

	cutilCheckMsg("sphCalDensity kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

#if USE_TEX
	cutilSafeCall(cudaUnbindTexture(dSortedPosTex));
	cutilSafeCall(cudaUnbindTexture(dCellStartTex));
	cutilSafeCall(cudaUnbindTexture(dCellEndTex));
#endif
}

/*!
 * p[eBN@̌vZ
 * @param[out] dNewDens p[eBNx
 * @param[out] dNewPres p[eBN
 * @param[in]  cell p[eBNObhf[^
 */
void CuSphNormal(float* dNrms, float* dDens, rxParticleCell cell)
{
	// MRK:CuSphNormal

	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalNormal<<< numBlocks, numThreads >>>((float4*)dNrms, dDens, cell);

	cutilCheckMsg("sphCalNormal kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

}

/*!
 * p[eBNɂ͂̌vZ(J[lĂяo)
 * @param[in] dDens p[eBNx
 * @param[in] dPres p[eBN
 * @param[out] dFrc p[eBNɂ
 * @param[in] cell p[eBNObhf[^
 * @param[in] dt ԃXebv
 */
void CuSphForces(float* dDens, float* dPres, float* dFrc, rxParticleCell cell, float dt)
{
#if USE_TEX
	cutilSafeCall(cudaBindTexture(0, dSortedPosTex, cell.dSortedPos, cell.uNumParticles*sizeof(float4)));
	cutilSafeCall(cudaBindTexture(0, dSortedVelTex, cell.dSortedVel, cell.uNumParticles*sizeof(float4)));
	cutilSafeCall(cudaBindTexture(0, dCellStartTex, cell.dCellStart, cell.uNumCells*sizeof(uint)));
	cutilSafeCall(cudaBindTexture(0, dCellEndTex, cell.dCellEnd, cell.uNumCells*sizeof(uint)));	
#endif

	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalForces<<< numBlocks, numThreads >>>(dDens, dPres, (float4*)dFrc, cell);

	cutilCheckMsg("calForcesSPH kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

#if USE_TEX
	cutilSafeCall(cudaUnbindTexture(dSortedPosTex));
	cutilSafeCall(cudaUnbindTexture(dSortedVelTex));
	cutilSafeCall(cudaUnbindTexture(dCellStartTex));
	cutilSafeCall(cudaUnbindTexture(dCellEndTex));
#endif
}

/*!
 * p[eBNʒuCx̍XV
 * @param[inout] pos p[eBNʒu
 * @param[inout] vel p[eBNx
 * @param[inout] velOld OXebṽp[eBNx
 * @param[in] frc p[eBNɂ
 * @param[in] dens p[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
void CuSphIntegrate(float* pos, float* vel, float* frc, float* dens, 
					float dt, uint nprts)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphIntegrate<<< numBlocks, numThreads >>>((float4*)pos, (float4*)vel, (float4*)frc, dens, 
											  dt, nprts);
	
	cutilCheckMsg("sphIntegrate kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
}


/*!
 * p[eBNʒuCx̍XV
 * @param[inout] pos p[eBNʒu
 * @param[inout] vel p[eBNx
 * @param[inout] velOld OXebṽp[eBNx
 * @param[in] frc p[eBNɂ
 * @param[in] dens p[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
void CuSphIntegrateWithPolygon(float* pos, float* vel, float* frc, float* dens, 
							   float* vrts, int* tris, int tri_num, float dt, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphIntegrateWithPolygon<<< numBlocks, numThreads >>>((float4*)pos, (float4*)vel, (float4*)frc, dens, 
											   (float3*)vrts, (int3*)tris, tri_num, dt, cell);
	
	cutilCheckMsg("sphIntegrate kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * Obh̖xZo
 * @param[out] dGridD Obh̖xl
 * @param[in] cell p[eBNObhf[^
 * @param[in] nx,ny Obh
 * @param[in] x0,y0 ObhŏW
 * @param[in] dx,dy Obh
 */
void CuSphGridDensity(float *dGridD, rxParticleCell cell, 
					  int nx, int ny, int nz, float x0, float y0, float z0, float dx, float dy, float dz)
{
#if USE_TEX
	cutilSafeCall(cudaBindTexture(0, dSortedPosTex, cell.dSortedPos, cell.uNumParticles*sizeof(float4)));
	cutilSafeCall(cudaBindTexture(0, dCellStartTex, cell.dCellStart, cell.uNumCells*sizeof(uint)));
	cutilSafeCall(cudaBindTexture(0, dCellEndTex, cell.dCellEnd, cell.uNumCells*sizeof(uint)));	
#endif

	uint3  gnum = make_uint3(nx, ny, nz);
	float3 gmin = make_float3(x0, y0, z0);
	float3 glen = make_float3(dx, dy, dz);

	int numcell = gnum.x*gnum.y*gnum.z;

	int threads = 128;
	dim3 grid((numcell+threads-1)/threads, 1, 1);
	if(grid.x > 65535){
		grid.y = (grid.x+32768-1)/32768;
		grid.x = 32768;
	}

	// J[ls
	sphCalDensityInGrid<<<grid, threads>>>(dGridD, cell, gnum, gmin, glen);

	cutilCheckMsg("Kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

#if USE_TEX
	cutilSafeCall(cudaUnbindTexture(dSortedPosTex));
	cutilSafeCall(cudaUnbindTexture(dCellStartTex));
	cutilSafeCall(cudaUnbindTexture(dCellEndTex));
#endif
}





//-----------------------------------------------------------------------------
// MARK:Anisotropic Kernel
//-----------------------------------------------------------------------------
/*!
 * J[lSʒu̍XVƏdݕtς̌vZ(J[l֐)
 * @param[out] dUpPos XVJ[lS
 * @param[out] dPosW dݕtσp[eBNW 
 * @param[in]  lambda ̂߂̒萔
 * @param[in]  h Ta
 * @param[in]  cell p[eBNObhf[^
 */
void CuSphCalUpdatedPosition(float* dUpPos, float* dPosW, float lambda, float h, rxParticleCell cell)
{
	// MRK:CuSphCalUpdatedPosition
#if USE_TEX
	cutilSafeCall(cudaBindTexture(0, dSortedPosTex, cell.dSortedPos, cell.uNumParticles*sizeof(float4)));
	cutilSafeCall(cudaBindTexture(0, dCellStartTex, cell.dCellStart, cell.uNumCells*sizeof(uint)));
	cutilSafeCall(cudaBindTexture(0, dCellEndTex, cell.dCellEnd, cell.uNumCells*sizeof(uint)));	
#endif

	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalUpdatedPosition<<< numBlocks, numThreads >>>((float4*)dUpPos, (float4*)dPosW, lambda, h, cell);

	cutilCheckMsg("sphCalUpdatedPosition kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

#if USE_TEX
	cutilSafeCall(cudaUnbindTexture(dSortedPosTex));
	cutilSafeCall(cudaUnbindTexture(dCellStartTex));
	cutilSafeCall(cudaUnbindTexture(dCellEndTex));
#endif
}

/*!
 * ʒuł̏dݕtψʒu̍ČvZcovariance matrix̌vZ
 * @param[out] dPosW dݕtσp[eBNW 
 * @param[out] dCMat Covariance Matrix
 * @param[in]  h Ta
 * @param[in]  cell p[eBNObhf[^
 */
void CuSphCalCovarianceMatrix(float* dPosW, float* dCMat, float h, rxParticleCell cell)
{
	// MRK:CuSphCalCovarianceMatrix
#if USE_TEX
	cutilSafeCall(cudaBindTexture(0, dSortedPosTex, cell.dSortedPos, cell.uNumParticles*sizeof(float4)));
	cutilSafeCall(cudaBindTexture(0, dCellStartTex, cell.dCellStart, cell.uNumCells*sizeof(uint)));
	cutilSafeCall(cudaBindTexture(0, dCellEndTex, cell.dCellEnd, cell.uNumCells*sizeof(uint)));	
#endif

	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalCovarianceMatrix<<< numBlocks, numThreads >>>((float4*)dPosW, (matrix3x3*)dCMat, h, cell);

	cutilCheckMsg("sphCalCovarianceMatrix kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

#if USE_TEX
	cutilSafeCall(cudaUnbindTexture(dSortedPosTex));
	cutilSafeCall(cudaUnbindTexture(dCellStartTex));
	cutilSafeCall(cudaUnbindTexture(dCellEndTex));
#endif
}

/*!
 * ْlɂŗLlvZ
 * @param[in]  dC Covariance Matrix
 * @param[in]  dPosW dݕtψʒu
 * @param[out] dEigen ŗLl
 * @param[out] dR ŗLxNg(]s)
 * @param[in]  numParticles p[eBN
 */
void CuSphSVDecomposition(float* dC, float* dPosW, float* dEigen, float* dR, 
						  float kr, int ne, float kn, float ks, uint numParticles)
{
	// MRK:CuSphCalTransformMatrix

	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(numParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphSVDecomposition<<< numBlocks, numThreads >>>((matrix3x3*)dC, (float4*)dPosW, (float3*)dEigen, (matrix3x3*)dR, 
													kr, ne, kn, ks, numParticles);

	cutilCheckMsg("sphCalTransformMatrix kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * ŗLlCŗLxNg(]s)ό`svZ
 * @param[in]  dEigen ŗLl
 * @param[in]  dR ŗLxNg(]s)
 * @param[out] dG ό`s
 * @param[in]  numParticles p[eBN
 */
void CuSphCalTransformMatrix(float* dEigen, float* dR, float *dG, uint numParticles)
{
	// MRK:CuSphCalTransformMatrix

	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(numParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalTransformMatrix<<< numBlocks, numThreads >>>((float3*)dEigen, (matrix3x3*)dR, (matrix3x3*)dG, numParticles);

	cutilCheckMsg("sphCalTransformMatrix kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
}


/*!
 * Obh̖xZo
 * @param[out] dGridD Obh̖xl
 * @param[in] cell p[eBNObhf[^
 * @param[in] nx,ny Obh
 * @param[in] x0,y0 ObhŏW
 * @param[in] dx,dy Obh
 */
void CuSphGridDensityAniso(float *dGridD, float *dG, rxParticleCell cell, 
						   int nx, int ny, int nz, float x0, float y0, float z0, float dx, float dy, float dz)
{
#if USE_TEX
	cutilSafeCall(cudaBindTexture(0, dSortedPosTex, cell.dSortedPos, cell.uNumParticles*sizeof(float4)));
	cutilSafeCall(cudaBindTexture(0, dCellStartTex, cell.dCellStart, cell.uNumCells*sizeof(uint)));
	cutilSafeCall(cudaBindTexture(0, dCellEndTex, cell.dCellEnd, cell.uNumCells*sizeof(uint)));	
#endif

	uint3  gnum = make_uint3(nx, ny, nz);
	float3 gmin = make_float3(x0, y0, z0);
	float3 glen = make_float3(dx, dy, dz);

	int numcell = gnum.x*gnum.y*gnum.z;

	int threads = THREAD_NUM;
	dim3 grid((numcell+threads-1)/threads, 1, 1);
	if(grid.x > 65535){
		grid.y = (grid.x+32768-1)/32768;
		grid.x = 32768;
	}

	// J[ls
	sphCalDensityAnisoInGrid<<<grid, threads>>>(dGridD, (matrix3x3*)dG, cell, gnum, gmin, glen);

	cutilCheckMsg("Kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

#if USE_TEX
	cutilSafeCall(cudaUnbindTexture(dSortedPosTex));
	cutilSafeCall(cudaUnbindTexture(dCellStartTex));
	cutilSafeCall(cudaUnbindTexture(dCellEndTex));
#endif
}


}   // extern "C"
